// TODO: After choosing fragment the header is not snippet correctly

$(document).ready(function() {
    prepareDatetimeFields();
    prepareDocument();
    prepareBody();
    prepareReplyButtons();
    prepareMessageContainers();
    prepareQuotedTexts();
    prepareAttachments();
    autosaveTimer = setupAutosaveTimer();
    undoSendTimer = setupUndoSendTimer();
    stateTimer = setupStateTimer();
    snippetSuggestionTimer = undefined; //setupSnippetSuggestionTimer();
    prepareDrafts();
    window.setTimeout(loadAttachments, 0);
    var selectedMessageContainer = $(".messagecontainer").first();
    selectMessageContainer(selectedMessageContainer);
    setInitialFocus(selectedMessageContainer);
});

var _autosave = true;

function setAutoSaveState(state) {
    _autosave = state;
}

function cleanUp() {
    clearInterval(autosaveTimer);
    clearInterval(undoSendTimer);
    clearInterval(stateTimer);
    clearInterval(snippetSuggestionTimer);
    InboxWindowController = undefined;
}

function getInboxWindowController() {
    return window.InboxWindowController;
}

function getUndoContext() {
    if (getInboxWindowController()) {
        return getInboxWindowController().undoContext();
    } else {
        return 0;
    }
}

function prepareDatetimeFields() {
    $(".datetime").cuteTime({ refresh: 10000 });
}

function doLoadNotReadyMailMessages() {
    $('.messagecontainer.notready').each(function(index) {
        var message = $(this);
        var mailMessageId = message.attr("mail_message_id");
        var isSelected = getSelectedMessageContainer() == message;
        $.get("/inbox/message", {mail_message_id : mailMessageId}, function(result) {
            var newMessage = $(result);
            message.replaceWith(newMessage);
            prepareMessageContainer(newMessage);
            loadAttachments();
            prepareDatetimeFields();
            if (isSelected) {
                selectMessageContainer(newMessage, false);
            }
        });
    });
}

function doMarkMessagesAsNotReady() {
    $(".messagecontainer").addClass("notready");
}

function doSaveDraftThenLoadConversation(conversation) {
    var drafts = $('div.draft.needs_save');
    if (drafts.length == 0) {
        return false;
    }
    drafts.each(function(index) {
        // only callback on last unsaved draft
        var callback_func = null;
        if (index == drafts.length - 1) {
            callback_func = function() {
                getInboxWindowController().loadConversation(conversation);
            };
        }
        save_draft($(this), callback_func);
    });
    return true;
}

function doInsertSnippetAtFocusedDraft(snippetId) {
    var draft = getFocusedDraft();
    if (!draft) {
        console.log("didn't find focused draft!");
        return;
    }
    insertSnippet(draft, snippetId, true);
}

function doReply() {
    var message = getSelectedMessageContainer();
    if (!message) {
        console.log("no message found.");
        return;
    }
    createReply(message.find('div.message'));
}

function validateReply() {
    if (isConversationDraft()) {
        return false;
    }
    var container = getSelectedMessageContainer();
    if (!container) return false;
    if (container.hasClass("has_draft")) return false;
    return true;
}

function doSelectNextMessage() {
    var nextMessage = getNextMessageContainer();
    if (nextMessage != null) {
        selectMessageContainer(nextMessage, true);
    }
}

function doSelectPreviousMessage() {
    var previousMessage = getPreviousMessageContainer();
    if (previousMessage != null) {
        selectMessageContainer(previousMessage, true);
    }
}

function doSendReply() {
    var draft = getFocusedDraft();
    if (draft.length == 0) {
        console.log("didn't find focused draft!");
        return;
    }
    sendReply(draft);
}

function doDiscardReply() {
    var draft = getFocusedDraft();
    if (draft.length == 0) {
        console.log("didn't find focused draft!");
        return;
    }
    discardReply(draft);
}

function doDiscardBodyFragment() {
    var bodyFragment = getFocusedBodyFragment();
    if (bodyFragment == null) {
        console.log("Didn't find focused body fragment!");
    }
    discardBodyFragment(bodyFragment);
}

function doAppendBodyFragment() {
    var bodyFragment = getFocusedBodyFragment();
    if (bodyFragment == null) {
        console.log("Didn't find focused body fragment!");
    }
    appendBodyFragmentAfter(bodyFragment);
}

function doAddFileAttachmentsToFocusedBodyFragment(fileListAsString) {
    var bodyFragment = getFocusedBodyFragment();
    if (bodyFragment == null) {
        console.log("Didn't find focused body fragment!");
    }
    addFileAttachmentsToBodyFragment(bodyFragment, fileListAsString);
}

function getFocusedBodyFragmentText() {
    var bodyFragment = getFocusedBodyFragment();
    if (bodyFragment == null) {
        return null;
    }
    var richtext = bodyFragment.find("section.richtext");
    var text = richtext.get(0).innerText;
    if (text && text.length > 0) {
        var lines = text.split("\n");
        if (lines.length > 0) {
            return lines[0];
        } else {
            return text;
        }
    } else {
        return null;
    }
}

function hasFocusedDraft() {
    if (getFocusedDraft().length) {
        return true;
    } else {
        return false;
    }
}

function hasFocusedBodyFragment() {
    var focusedFragments = getFocusedBodyFragment();
    if (!focusedFragments || focusedFragments.length == 0) {
        return false;
    } else {
        return true;
    }
}

function validateAppendBodyFragment() {
    return hasFocusedBodyFragment();
}

function getFocusedBodyFragmentId() {
    var fragment = getFocusedBodyFragment();
    if (!fragment) {
        return null;
    }
    return parseInt(fragment.attr("body_fragment_id"));
}

function doUndoSend() {
    var messageContainer = getSelectedMessageContainer();
    if (!messageContainer) {
        return;
    }
    undoSend(messageContainer);
}

function doAcceptSnippetSuggestion() {
    insertSnippetSuggestion();
}

function validateAcceptSnippetSuggestion() {
    var popover = getSnippetSuggestionPopover();
    if (!popover.is(":visible")) {
        return false;
    }
    var snippetId = popover.find(".snippet_suggestion_content").attr("snippet_id");
    if (!snippetId) {
        return false;
    }
    return true;
}

function doSwitchDraftAddrsToReplyAll() {
    switchReplyAddrs(getFocusedDraft(), 'reply_all');
}

function doSwitchDraftAddrsToSingle() {
    switchReplyAddrs(getFocusedDraft(), 'single');
}

function validateSwitchDraftAddrsToSingle() {
    var draft = getFocusedDraft();
    if (!draft) {
        return false;
    }
    var link = draft.find(".switch_to_single_reply");
    return link.is(":visible");
}

function validateSwitchDraftAddrsToReplyAll() {
    var draft = getFocusedDraft();
    if (!draft) {
        return false;
    }
    var link = draft.find(".switch_to_reply_all");
    return link.is(":visible");
}

function hasPreviousMessage() {
    return getPreviousMessageContainer() != null;
}

function hasNextMessage() {
    return getNextMessageContainer() != null;
}

function doExpandSelectedMessage() {
    expandMessageContainer(getSelectedMessageContainer());
}

function doCollapseSelectedMessage() {
    collapseMessageContainer(getSelectedMessageContainer());
}

function doCollapseAllMessages() {
    $(".messagecontainer").each(function(index) {
        var messageContainer = $(this);
        if (canCollapseSelectedMessage(messageContainer)) {
            collapseMessageContainer(messageContainer);
        }
    });
    moveMessageSelectionBackgroundTo(getSelectedMessageContainer());
}

function doExpandAllMessages() {
    $(".messagecontainer").each(function(index) {
        var messageContainer = $(this);
        if (canExpandMessageContainer(messageContainer)) {
            expandMessageContainer(messageContainer);
        }
    });
    moveMessageSelectionBackgroundTo(getSelectedMessageContainer());
}

function doClearTextFormatting() {
    if (hasFocusedBodyFragment()) {
        document.execCommand("removeformat");
    }
}

function canCollapseSelectedMessage() {
    return canCollapseMessageContainer(getSelectedMessageContainer());
}

function canCollapseMessageContainer(messageContainer) {
    if ($(".thread").hasClass("notpingpong")) {
        return false;
    }
    if (messageContainer.hasClass("has_draft")) {
        return false;
    }
    if (messageContainer.hasClass("is_inbox")) {
        return false;
    }
    return !(messageContainer.hasClass("collapsed"));
}

function canExpandSelectedMessage() {
    return canExpandMessageContainer(getSelectedMessageContainer());
}

function canExpandMessageContainer(messageContainer) {
    if ($(".thread").hasClass("notpingpong")) {
        return false;
    }
    if (messageContainer.hasClass("has_draft")) {
        return false;
    }
    if (messageContainer.hasClass("is_inbox")) {
        return false;
    }
    return messageContainer.hasClass("collapsed");
}

function isSelectedMessageExpanded() {
    var selectedMessage = getSelectedMessageContainer();
    return !(selectedMessage.hasClass("collapsed"));
}

function validateExpandAllMessages() {
    if ($(".thread").hasClass("notpingpong")) {
        return false;
    }
    var count = 0
    $(".messagecontainer").each(function(index) {
        var messageContainer = $(this);
        if (canExpandMessageContainer(messageContainer)) {
            count += 1;
        }
    });
    return count > 0;
}

function validateCollapseAllMessages() {
    if ($(".thread").hasClass("notpingpong")) {
        return false;
    }
    var count = 0
    $(".messagecontainer").each(function(index) {
        var messageContainer = $(this);
        if (canCollapseMessageContainer(messageContainer)) {
            count += 1;
        }
    });
    return count > 0;
}

function hasSelectedMessage() {
    var message = getSelectedMessageContainer();
    return message.length != 0;
}

function getSelectedMessageId() {
    var message = getSelectedMessageContainer();
    if (message.length) {
        return parseInt(message.attr("message_id"));
    } else {
        return null;
    }
}

function getPreviousMessageContainer() {
    var prevContainer = getSelectedMessageContainer().prev();
    if (prevContainer.is("div.messagecontainer")) {
        return prevContainer;
    }
    return null;
}

function prepareDocument() {
    $(document).scroll();
}

function prepareBody() {
    var body = $('body');
    prepareBodyAsDropZone(body);
}

function prepareBodyAsDropZone(body) {
    body = $('html');
    body.bind('dragstart', function(event) {
        return false;
    });
    body.bind('dragenter', function(event) {
        body.addClass('dragover');
        return false;
    });
    body.bind('dragleave', function(event) {
        body.removeClass('dragover');
        return false;
    });
    body.bind('dragover', function(event) {
        var dt = event.originalEvent.dataTransfer;
        if (hasFocusedBodyFragment()) {
            dt.effectAllowed = "all";
            dt.dropEffect = "copy";
        } else {
            dt.effectAllowed = "none";
            dt.dropEffect = "none";
        }
        return false;
    });
    body.bind('drop', function(ev) {
        var bodyFragment = getFocusedBodyFragment();
        if (!bodyFragment) {
            return;
        }
        var dataTransfer = ev.originalEvent.dataTransfer;
        var fileList = dataTransfer.getData('text/uri-list');
        if (!fileList) {
            console.log("[WARN] didn't find any files to attach.");
            body.removeClass('dragover');
            return true;
        }
        addFileAttachmentsToBodyFragment(bodyFragment, fileList);
        body.removeClass('dragover');
        return false;
    });
}


function preventDropsOnBody(body) {
    body.bind('dragenter', function(event) {
        return false;
    });
    body.bind('dragleave', function(event) {
        return false;
    });
    body.bind('dragover', function(event) {
        return false;
    });
    body.bind('drop', function(ev) {
        return false;
    });
}

function prepareMessageContainers() {
    $("div.messagecontainer").each(function(inbox) {
        prepareMessageContainer($(this));
    });
}

function prepareMessageContainer(messageContainer) {
    messageContainer.click(function(event) {
        selectMessageContainer($(event.target).parents('div.messagecontainer'), false);
        if (event.altKey && validateReply()) {
            var message = messageContainer.find("div.message");
            if (message.length != 0) {
                createReply(messageContainer.find("div.message"));
                event.preventDefault();
            }
        }
    });
    messageContainer.find("div.collapsed_message").each(function(index) {
        var collapsedMessage = $(this);
        prepareCollapsedMessage(collapsedMessage, messageContainer);
    });
    messageContainer.find("div.message").each(function(index) {
        var messageElement = $(this);
        prepareMessage(messageElement);
    });
    messageContainer.find("a.contactlink").each(function(index) {
        var contactLink = $(this);
        prepareContactLink(contactLink)
    });
}

function prepareContactLink(contactLink) {
    contactLink.attr("href", "");
    contactLink.click(function(event) {
        var contactLink = $(this);
        getInboxWindowController().showContactInfoForId(parseInt(contactLink.attr("contact_id")));
    });
}

function prepareCollapsedMessage(collapsedMessage, messageContainer) {
    collapsedMessage.find("div.message_header").click(function(event) {
        selectMessageContainer(messageContainer, false);
        return false;
    });
    collapsedMessage.find(".selection_indicator").click(function(event) {
        expandMessageContainer(messageContainer);
        selectMessageContainer(messageContainer, false);
        return false;
    });
}

function collapseMessageContainer(messageContainer) {
    messageContainer.addClass("collapsed");
    moveMessageSelectionBackgroundTo(messageContainer);
}

function expandMessageContainer(messageContainer) {
    messageContainer.removeClass("collapsed");
    moveMessageSelectionBackgroundTo(messageContainer);
}

function prepareMessage(messageElement) {
    prepareDraftUndoSend(messageElement);
    var messageContainer = messageElement.parents("div.messagecontainer");
    messageElement.find("div.message_header").click(function(event) {
        selectMessageContainer(messageContainer, false);
        return false;
    });
    messageElement.find("div.message_header .selection_indicator").click(function(event) {
        selectMessageContainer(messageContainer, false);
        collapseMessageContainer(messageContainer);
        return false;
    });
}

function getNextMessageContainer() {
    var nextContainer = getSelectedMessageContainer().next();
    if (nextContainer.is("div.messagecontainer")) {
        return nextContainer;
    }
    return null;
}

function getSelectedMessageContainer() {
    return $("div.messagecontainer.selected");
}

function selectMessageContainer(messageContainer, scroll) {
    hideSnippetSuggestion();
    getSelectedMessageContainer().removeClass("selected");
    messageContainer.addClass("selected");
    computeThreadPingPong();
    moveMessageSelectionBackgroundTo(messageContainer);
    if (scroll) {
        scrollDivIntoView(messageContainer);
    }
}

function setInitialFocus(messageContainer) {
    if (messageContainer.hasClass("has_draft")) {
        var draft = messageContainer.find(".draft");
        var message_header = draft.find(".message_header");
        var focused = false;
        if (message_header.hasClass("contacts_editable") && hasEmptyToAddr(draft)) {
            draft.find(".message_header .to_addr input").focus();
            focused = true;
        } else if (message_header.hasClass("subject_editable")) {
            subjectInput = draft.find(".message_header input[name='subject']");
            if (subjectInput.val().length == 0) {
                subjectInput.focus();
                focused = true;
            }
        }
        focusBodyFragment(draft.find(".body_fragment.initial_focus"), !focused);
    }
}


function moveMessageSelectionBackgroundTo(messageContainer) {
    var selectionBackground = $("div.message_selection_background");
    if ($("div.messagecontainer").length <= 1 || messageContainer == null || messageContainer.is(":visible") == false) {
        selectionBackground.hide();
        return;
    }
    var height = messageContainer.height()
    var top = messageContainer.offset().top;
    selectionBackground.height(height + 10);
    selectionBackground.offset({top: top - 6, left: 0});
    selectionBackground.show();
}


function getFocusedBodyFragment() {
    var draft = getFocusedDraft();
    if (!draft) {
        return null;
    }
    var result = draft.find('div.body_fragment.focus');
    if (result.length == 0) {
        return null;
    }
    return result;
}

function getFocusedBodyFragmentRichTextHTML() {
    var fragment = getFocusedBodyFragment();
    if (!fragment) {
        return null;
    }
    var richtext = fragment.find('section.richtext');
    return richtext.html();
}

function getFocusedDraft() {
    return $('div.messagecontainer.selected div.draft');
}

/* Returns the absolute position of the focused draft */
function getFocusedBodyFragmentPosition() {
    return getVisiblePosition(getFocusedBodyFragment());
}

function getDraftForm(draft) {
    return draft.find('form.send_reply');
}

function discardBodyFragment(bodyFragment) {
    var draft = bodyFragment.parents("div.draft");
    save_draft(draft, function() {
        var bodyFragmentId = bodyFragment.attr("body_fragment_id");
        var url = "/inbox/draft/body_fragment/" + bodyFragmentId + "/discard/" + getUndoContext();
        $.post(url, function(result) {
            replaceDraft(draft, $(result));
        });
    });
}

function appendBodyFragmentAfter(bodyFragment) {
    var draft = bodyFragment.parents("div.draft");
    save_draft(draft, function() {
        var bodyFragmentId = bodyFragment.attr("body_fragment_id");
        var url = "/inbox/draft/body_fragment/" + bodyFragmentId + "/append/" + getUndoContext();
        $.post(url, function(result) {
            replaceDraft(draft, $(result));
        });
    });
}


function replaceDraft(oldDraft, newDraft, callback_func) {
    hideSnippetSuggestion();
    oldDraft.replaceWith(newDraft);
    prepareDraft(newDraft);
    prepareQuotedTexts();
    var focusedFragment = newDraft.find("div.body_fragment.initial_focus");
    focusBodyFragment(focusedFragment, true);
    if (focusedFragment.length > 0) {
        scrollDivIntoView(focusedFragment, callback_func);
    } else {
        scrollDivIntoView(newDraft, callback_func);
    }
    loadAttachments();
    prepareAttachments();
}

function insertSnippet(draft, snippetId) {
    var bodyFragment = getFocusedBodyFragment();
    var draft = bodyFragment.parents('div.draft');
    var bodyFragmentId = bodyFragment.attr('body_fragment_id');
    save_draft(draft, function() {
        $.get("/inbox/draft/body_fragment/" + bodyFragmentId + "/insert_snippet/" + snippetId + "/" + getUndoContext(), function(result) {
            var new_draft = $(result);
            replaceDraft(draft, new_draft);
        });
    });
}

function validateInsertSentMessageAtFocusedDraft() {
    if ($(".body_fragment section:focus").length == 0) {
        return false;
    }
    return getFocusedBodyFragment() != null;
}

function setupAutosaveTimer() {
    return window.setInterval(auto_save_drafts, 2000);
}

function setupSnippetSuggestionTimer() {
    return window.setInterval(checkSnippetSuggestion, 300);
}

function auto_save_drafts() {
    $('div.draft.needs_save').each(function(event) {
        if (_autosave) {
            var draft = $(this);
            save_draft(draft);
        } else {
            console.log("ignoring auto save for " + draft);
        }
    });
}

function setupUndoSendTimer() {
    return window.setInterval(checkUndoSendLinks, 2000);
}

function hasEmptyToAddr(draft) {
    return draft.find(".message_header .to_addr .tagedit-listelement-old:not(.tagedit-listelement-deleted)").length == 0;
}

function setupStateTimer() {
    return window.setInterval(function() {
        var hasFocusedDraftState = hasFocusedDraft();
        var messageContainer = getSelectedMessageContainer();
        var undo_link = messageContainer.find("a.undo_send");
        var canUndoSend = undo_link.length != 0;
        var isSending = messageContainer.hasClass("sending");

        function computeHasValidToAddress() {
            var draft = getFocusedDraft();
            var error_msg = draft.find(".message_header .contact_controls .error_msg");
            var our_error_message = "Please enter one email address.";
            if (!draft) {
                if (error_msg.text() == our_error_message) {
                    error_msg.text("");
                }
                return false;
            }
            if (draft.find(".message_header.contacts_editable").length == 0) {
                if (error_msg.text() == our_error_message) {
                    error_msg.text("");
                }
                return true;
            }
            if (draft.find(".message_header .tagedit-list").length == 0) {
                if (error_msg.text() == our_error_message) {
                    error_msg.text("");
                }
                return true;
            }
            if (!hasEmptyToAddr(draft)) {
                if (error_msg.text() == our_error_message) {
                    error_msg.text("");
                }
                return true;
            }
            if (error_msg.text() == "") {
                error_msg.text("Please enter one email address.");
            }
            return false;
        }

        var wc = getInboxWindowController();
        if (wc) {
            wc.setIsSending(isSending)
            wc.setCanReply(validateReply());
            wc.setHasFocusedDraft(hasFocusedDraftState);
            wc.setCanUndoSend(canUndoSend);
            wc.setCanSend(hasFocusedDraftState && !isSending && computeHasValidToAddress());
        }
    }, 200);
}

function checkUndoSendLinks() {
    $("a.undo_send").each(function(event) {
        var link = $(this);
        var draftId = link.attr("draft_id");
        $.getJSON("/inbox/draft/" + draftId + "/state", function(result) {
            if (result.state == 2) { // processing
                link.fadeOut();
            } else if (result.state > 2) { // processed
                var message = link.parents("div.message");
                var date = message.find("div.message_header span.datetime");
                date.text(result.date);
                date.cuteTime({refresh: 10000});
                link.remove();
            }
        });
    });
}

function save_draft(draft, callback_func) {
    draft.removeClass("needs_save");
    var formElement = draft.find('form.send_reply');
    if (!formElement) {
        getInboxWindowController().showWarningMessage("Could not save message draft.", "Reason: Error finding UI element. If the problem persists, please contact support.");
        console.log("[ERROR] couldn't find formelement 'form.send_reply'");
        if (callback_func) {
            callback_func();
        }
        return;
    }
    formElement.parents('div.draft').find('div.body_fragment').each(function(index) {
        var bodyFragment = $(this);
        var richText = bodyFragment.find("section.richtext");
        var input = bodyFragment.find("input[name$='body']");
        if (richText.html() != input.val()) {
            input.val(richText.html());
        }
    });
    $.post("/inbox/draft/"+ draft.attr("draft_id") + "/save", formElement.serialize(), function(result) {
        if (result.status != "OK") {
            console.log("[WARN] could not save draft!");
            // getInboxWindowController().showWarningMessage("Could not save message draft.", "Reason: Couldn't save to database. If the problem persists, please contact support.");
        }
        if (result.greeting) {
            draft.find(".body_fragment.greeting section").html(result.greeting);
        }
        if (result.signature) {
            draft.find(".body_fragment.signature section").html(result.signature);
        }
        if (result.message_header) {
            var newHeader = $(result.message_header);
            prepareContactsEditable(newHeader);
            draft.find(".to_addr .tagedit-listelement-old").replaceWith(newHeader.find(".to_addr .tagedit-listelement-old"));
            draft.find(".cc_addr .tagedit-listelement-old").replaceWith(newHeader.find(".cc_addr .tagedit-listelement-old"));
            draft.find(".bcc_addr .tagedit-listelement-old").replaceWith(newHeader.find(".bcc_addr .tagedit-listelement-old"));
        }
        if (callback_func) {
            callback_func();
        }
    }, 'json');
}

function _insertDraft(draft, messageContainer, callback_func) {
    draft.hide();
    draft.appendTo(messageContainer.find('div.drafts'));
    prepareQuotedTexts();
    messageContainer.addClass('has_draft');
    messageContainer.children('div.message').hide();
    messageContainer.removeClass("answered").addClass("unanswered");
    draft.fadeIn("normal", function() {
        prepareDraft(draft);
        setInitialFocus(messageContainer);
        if (callback_func) {
            callback_func();
        }
    });
    prepareAttachments();
    loadAttachments();
}

function prepareReplyButtons() {
    $("form.createreply.unbound").each(function(index) {
        var formElement = $(this);
        formElement.removeClass("unbound");
        formElement.submit(function(event) {
            event.preventDefault();
            var messageElement = formElement.parents('div.message');
            createReply(messageElement);
        });
    });
}

function createReply(messageElement) {
    var formElement = messageElement.find('form.createreply');
    var messageId = formElement.attr("message_id");
    var messageContainer = formElement.parents('div.messagecontainer');
    expandMessageContainer(messageContainer);
    $.get("/inbox/reply/" + messageId, function(result) {
        var draft = $(result);
        _insertDraft(draft, messageContainer);
    });
}


function scrollDivIntoView(draft, callback_func) {
    var scrollTop = $('body').scrollTop();
    var draftHeight = draft.outerHeight();
    var windowHeight = $(window).height();
    var positionTop = draft.position().top;
    var lowerPosition = positionTop - scrollTop + draftHeight;
    if (positionTop < scrollTop) {
        $("body").animate({
            scrollTop : positionTop // - 8
        }, 200, callback_func);
    } else if (lowerPosition > windowHeight) {
        if (windowHeight < draftHeight) {
            $("body").animate({
                scrollTop : positionTop
            }, 200, callback_func);
        } else {
            var scrollOffset = lowerPosition - windowHeight;
            $("body").animate({
                scrollTop : scrollTop + scrollOffset + 16
            }, 200, callback_func);
        }
    } else {
        if (callback_func) {
            callback_func();
        }
    }
}

function prepareDrafts() {
    $("div.draft.unbound").each(function(index) {
        var draftElement = $(this);
        draftElement.parents("div.messagecontainer").children("div.message").hide();
        prepareDraft(draftElement);
    });
}

function prepareDraft(draftElement) {
    draftElement.removeClass("unbound");
    draftElement.find('.sending').hide();
    var formElement = draftElement.find("form.send_reply");
    prepareFromIdentity(draftElement, formElement);
    prepareContactsEditable(draftElement);
    prepareDraftBodyFragments(draftElement, formElement);
    prepareSearchButton(draftElement, formElement);
    prepareDraftForm(formElement);
    prepareDraftSendButton(formElement);
    prepareDraftDiscardButton(formElement, draftElement);
    draftElement.find("a.switch_to_single_reply").each(function(index) {
        prepareSwitchToSingleReply($(this));
    });
    draftElement.find("a.switch_to_reply_all").each(function(index) {
        prepareSwitchToReplyAll($(this));
    });
    draftElement.find("a.contactlink").each(function(index) {
        var contactLink = $(this);
        prepareContactLink(contactLink)
    });
}

function prepareFromIdentity(draftElement, formElement) {
    draftElement.find("select.from_identity").change(function(event) {
        var focused_fragment_id = getFocusedBodyFragmentId();
        var draft_id = draftElement.attr("draft_id");
        var identity = $(this).val();
        save_draft(draftElement, function () {
            var url = "/inbox/draft/" + draft_id + "/change_identity/" + identity + "/focus/" + focused_fragment_id;
            $.post(url, function (result) {
                replaceDraft(draftElement, $(result));
            });
        });
    });
}

function prepareContactsEditable(draftElement) {
    var draftId = draftElement.attr("draft_id");

    function prepareContactsGroup(type) {
        function enableTagedit() {
            draftElement.find("input." + type + "-contact-editable").tagedit({
                autocompleteURL:"/inbox/draft/" + draftId + "/autocomplete_contacts/" + type,
                breakKeyCodes:[ 13, 9 ],
                allowEdit:false,
                allowDelete:false,
                animSpeed:200
            });
        }
        var enablerLink = draftElement.find(".add_" + type);
        if (draftElement.find("." + type + "_addr").hasClass("empty")) {
            enablerLink.show().removeClass("empty");
        } else {
            enableTagedit();
            draftElement.find("." + type + "_addr").attr("style", "opacity: 100;");
        }
        enablerLink.click(function (event) {
            $(this).hide();
            draftElement.find("." + type + "_addr.label").show().attr("style", "display: inline;");
            draftElement.find("." + type + "_addr.value").show().attr("style", "display: table;");
            enableTagedit();
        });
    }

    prepareContactsGroup("to");
    prepareContactsGroup("cc");
    prepareContactsGroup("bcc");

    $("body").bind("contacts_changed", function(event) {
        setTimeout(function() {
            save_draft(draftElement);
        }, 200);
    });

    $("body").bind("contact_address_error", function(event, email) {
        draftElement.find(".contact_controls .error_msg").text("'" + email + "' is not a valid email address.")
    });

    $("body").bind("contact_clear_address_error", function(event) {
        draftElement.find(".contact_controls .error_msg").text("");
    });

    draftElement.find(".edit_contacts").click(function(event) {
        draftElement.find("input[name='are_contacts_editable']").val(1);
        save_draft(draftElement, function() {
            $.get("/inbox/draft/" + draftId + "/header", function(result) {
                replaceMessageHeader(draftElement, $(result));
            });
        });
    });

    draftElement.find("a.edit_subject").click(function(event) {
        $(this).hide();
        draftElement.find(".message_header .subject").show();
        draftElement.find("input[name='is_subject_editable']").val(1);
    });
    function changeHandler() {
        draftElement.addClass("needs_save");
    }
    var subjectElement = draftElement.find(".message_header .subject input");
    subjectElement
        .bind('keypress', changeHandler)
        .bind('keydown', changeHandler)
        .bind('change', changeHandler)
        .bind('blur', changeHandler);
}

function replaceMessageHeader(draftElement, newHeader) {
    draftElement.find(".message_header").replaceWith(newHeader);
    setTimeout(function (event) {
        prepareFromIdentity(draftElement);
        prepareContactsEditable(draftElement);
    }, 0);
}

function switchReplyAddrs(draft, type) {
    var focused_fragment_id = getFocusedBodyFragmentId();
    var draft_id = draft.attr("draft_id");
    save_draft(draft, function() {
        var url = "/inbox/draft/" + draft_id + "/switch_addrs/" + type + "/focus/" + focused_fragment_id;
        $.post(url, function(result) {
            replaceDraft(draft, $(result));
        });
    });
}

function prepareSwitchToSingleReply(link) {
    link.click(function(event) {
        var draft = link.parents(".draft");
        switchReplyAddrs(draft, "single");
    });
}

function prepareSwitchToReplyAll(link) {
    link.click(function(event) {
        var draft = link.parents(".draft");
        switchReplyAddrs(draft, "reply_all");
    });
}

function prepareDraftBodyFragments(draftElement, formElement) {
    formElement.find('div.body_fragment').each(function(index) {
        bodyFragment = $(this);
        prepareBodyFragment(draftElement, formElement, bodyFragment);
    });
}

function getSnippetSuggestionPopover() {
    return $(".snippet_suggestion_popover");
}
function hideSnippetSuggestion() {
    getSnippetSuggestionPopover().fadeOut("fast");
}

function insertSnippetSuggestion() {
    if (!validateAcceptSnippetSuggestion()) {
        return;
    }
    var snippetId = getSnippetSuggestionPopover().find(".snippet_suggestion_content").attr("snippet_id");
    hideSnippetSuggestion();
    insertSnippet(getFocusedDraft(), snippetId);
}

previousText = null;

function checkSnippetSuggestion() {
    var bodyFragment = getFocusedBodyFragment();
    if (!bodyFragment) {
        return;
    }
    var richtext = bodyFragment.find('section.richtext');
    var text = richtext.text();
    if (text != previousText) {
        console.log("checking text '" + text + "'")
        previousText = text;
        var url = "/inbox/draft/body_fragment/" + bodyFragment.attr("body_fragment_id") + "/suggest_snippet";
        var data = {
            input: text
        };
        var successFunc = function(result, textStatus, request) {
            if (result.search("Response404") != -1) {
                hideSnippetSuggestion();
                return;
            }
            var popover = getSnippetSuggestionPopover();
            var content = $(result);
            popover.find(".content").html(content);

            var richtextOffset = richtext.offset();
            popover.css("top", richtextOffset.top + richtext.height() + 40);
            popover.css("left", richtextOffset.left + richtext.width() * 0.05)
            popover.css("width", richtext.width() * 0.9);
            var popoverContent = popover.find(".snippet_suggestion_content");
            popover.find(".insert_suggested_snippet").click(function(event) {
                doAcceptSnippetSuggestion();
            });
            popoverContent.click(function(event) {
                doAcceptSnippetSuggestion();
            });
            popoverContent.addClass("fade_out");

            popover.fadeIn("fast", function() {
                if (popoverContent.height() < 200) {
                    popoverContent.removeClass("fade_out");
                }
            });
        };
        $.ajax({type: 'POST', url: url, data: data, error: hideSnippetSuggestion, success: successFunc});
    }
}

function prepareBodyFragment(draftElement, formElement, bodyFragment) {
    var body = bodyFragment.find('input[name$="body"]');
    var richtext = bodyFragment.find('section.richtext');
    prepareRichText(bodyFragment, richtext);
    function changeHandler(event) {
        var html = richtext.html();
        body.val(html);
        draftElement.addClass("needs_save");
    }

    function keyPressHandler(event) {
        changeHandler(event);
    }

    function keyDownHandler(event) {
    }

    function blurHandler(event) {
        changeHandler(event);
        hideSnippetSuggestion();
    }

    function focusHandler(event) {
        if (bodyFragment.hasClass("initial_focus")) {
            bodyFragment.removeClass("intial_focus");
        }
    }

    function pasteHandler(event) {
        var clipboard = event.originalEvent.clipboardData;
        var data = clipboard.getData('Image');
    }

    richtext
        .bind('keypress', keyPressHandler)
        .bind('keydown', keyDownHandler)
        .bind('change', changeHandler)
        .bind('focus', focusHandler)
        .bind('blur', blurHandler)
        .html(body.val());
    bodyFragment.find('div.quoted').each(function(index) {
        var quoted = $(this);
        prepareDraftQuoted(draftElement, formElement, quoted);
    });
    bodyFragment.find("a.searchsnippet").click(function(event) {
        event.preventDefault();
        searchSnippetForBodyFragment(bodyFragment);
    });
    bodyFragment.find("a.new_attachment").click(function(event) {
        event.preventDefault();
        getInboxWindowController().newAttachment();
    });
    bodyFragment.find("div.append_fragment_zone, a.addfragment").click(function(event) {
        event.preventDefault();
        appendBodyFragmentAfter(bodyFragment);
    });
    bodyFragment.find("a.removefragment").click(function(event) {
        event.preventDefault();
        discardBodyFragment(bodyFragment);
    });
}

function prepareRichText(bodyFragment, richtext) {
    richtext.focus(function(event) {
        focusBodyFragment(bodyFragment, false);
    });
}

function prepareDraftForm(formElement) {
    formElement.submit(function(event) {
        event.preventDefault();
    });
}

function prepareDraftUndoSend(messageElement) {
    messageElement.find('a.undo_send.unbound').removeClass("unbound").click(function(event) {
        var link = $(this);
        undoSend(link.parents("div.messagecontainer"));
    });
}

function undoSend(messageContainer) {
    var link = messageContainer.find("a.undo_send");
    if (!link) {
        console.log("[ERROR] could not find undo link.");
        return;
    }
    var targetMessageId = link.attr("message_id");
    var draftId = link.attr("draft_id");
    var targetMessageContainer = $('div.messagecontainer[message_id="' + targetMessageId + '"]');
    $.get("/inbox/draft/" + draftId + "/undo_send", function(result) {
        messageContainer.remove();
        _insertDraft($(result), targetMessageContainer);
        selectMessageContainer(targetMessageContainer, true);
    });
}

function prepareSearchButton(draftElement, formElement) {
    var draftId = draftElement.attr("draft_id");
    draftElement.find('a.search').click(function(event) {
        formElement.find('section.richtext').addClass('focus');
        event.preventDefault();
        getInboxWindowController().openSearchPanel();
    });
}

function searchSnippetForBodyFragment(bodyFragment) {
    getInboxWindowController().openSearchPanel();
}

function prepareDraftSendButton(formElement) {
    var draft = formElement.parents('div.draft');
    formElement.find('button.send_draft').click(function(event) {
        event.preventDefault();
        sendReply(draft);
    });
}

function sendReply(draft) {
    var formElement = getDraftForm(draft);
    formElement.find('.draft_actions').hide();
    formElement.find('.sending').show();
    var messageContainer = draft.parents("div.messagecontainer");
    messageContainer.addClass("sending");
    scrollDivIntoView(draft);
    save_draft(draft, function() {
        $.post("/inbox/draft/send/" + getUndoContext(), formElement.serialize(), function(result) {
            var content = $(result);
            draft.remove();
            messageContainer.removeClass("sending");
            content.insertBefore(messageContainer);
            messageContainer.removeClass("has_draft");
            content.hide();
            prepareQuotedTexts();
            prepareMessageContainer(content);
            messageContainer.removeClass("unanswered").addClass("answered");
            var message = messageContainer.find("div.message");
            message.show(0, function() {
                selectMessageContainer(content, true);
                content.fadeIn(1000);
            });
        });
    });
}

function prepareDraftDiscardButton(formElement, draftElement) {
    formElement.find('button.discard_draft').click(function() {
        discardReply(draftElement);
    });
}

function discardReply(draftElement) {
    var draftId = draftElement.attr("draft_id");
    $.get("/inbox/draft/discard/" + draftId + "/" + getUndoContext(), function(result) {
        if (result == "OK") {
            var messageContainer = draftElement.parents('div.messagecontainer');
            messageContainer.removeClass("has_draft");
            // change style before slideUp, because the reply button
            // would be hidden
            draftElement.hide();
            draftElement.remove();
            var messageElement = messageContainer.children("div.message");
            messageElement.fadeIn(function() {
                draftElement.remove();
            });
            selectMessageContainer(messageContainer, false);
        }
    });
}

function addFileAttachmentsToBodyFragment(bodyFragment, fileList) {
    var attachments = bodyFragment.find('div.attachments');
    $.post("/inbox/draft/body_fragment/" + bodyFragment.attr("body_fragment_id") + "/attachment/add", {'files' : fileList},
        function(result) {
            var attachments = $(result).children();
            appendAttachments(bodyFragment, attachments);
        });
}

function appendAttachments(bodyFragment, attachments) {
    var dropzone = bodyFragment.find("div.dropzone");
    attachments.hide();
    attachments.insertBefore(dropzone);
    attachments.fadeIn("normal", function() {
        scrollDivIntoView(bodyFragment);
        prepareAttachments();
    });
}

function prepareAttachments() {
    $('div.attachments')
        .each(function(index) {
            $(this)
                .find('div.draftattachment.unbound')
                .each(function(index) {
                    var attachmentElement = $(this);
                    attachmentElement.removeClass("unbound");
                    var attachmentId = attachmentElement.attr("attachment_id");
                    attachmentElement
                        .find('a.discard')
                        .each(function(index) {
                            $(this).click(function(event) {
                                $.post("/inbox/draft/attachment/" + attachmentId + "/remove", function(result) {
                                    attachmentElement.fadeOut("normal", function() {
                                        attachmentElement.remove();
                                    });
                                });
                                return false;
                            });
                        });
                    attachmentElement
                        .find('select.optimization_type')
                        .each(function(index) {
                            var optimizationType = $(this);
                            optimizationType.change(function(event) {
                                var optimizationTypeValue = optimizationType.val();
                                $.post("/inbox/draft/attachment/"
                                    + attachmentId
                                    + "/change_optimization_type/"
                                    + optimizationTypeValue,
                                    function(result) {
                                        attachmentElement.find('span.name').text(result);
                                    });
                            });
                        });
                });
        });
}

function prepareQuotedTexts() {
    var quoteElements = 'blockquote:not(.quoted_parent)';
    $("div.message, div.quoted").each(function(index) {
        var container = $(this);
        if (container.find('div.gmail_quote > *').not('blockquote, div.gmail_quote').length == 0) {
            quoteElements = '.gmail_quote:not(.quoted_parent), ' + quoteElements;
        }
        container.find(quoteElements).each(function(index) {
            var element = $(this);
            if (element.parents("blockquote").length == 0) {
                element.addClass("quoted_parent");
                element.addClass("quoted_hidden");
                var link = $('<br/><div class="toggle_quoted on"><span>Show Quoted Text</span></div>');
                link.insertBefore(element);
                link.click(function(event) {
                    event.preventDefault();
                    if (element.hasClass("quoted_hidden")) {
                        link.removeClass("on");
                        element.removeClass("quoted_hidden");
                    } else {
                        link.addClass("on");
                        element.addClass("quoted_hidden");
                    }
                });
                element.click(function(event) {
                    if (event.target.tagName != 'A') {
                        element.addClass("quoted_hidden");
                        link.addClass("on");
                        event.preventDefault();
                    }
                });
                element.find("a").click(function(event) {
                    var theTarget = $(this);
                    window.location.href = theTarget.attr("href");
                });
            }
        });
    });
}

function loadAttachments() {
    $("div.attachments.needsloading").each(function(index) {
        var attachmentElement = $(this);
        $.get("/inbox/attachments", {
            message_id : $(this).attr("message_id")
        }, function(result) {
            attachmentElement.html(result);
            attachmentElement.removeClass("needsloading");
            attachmentElement.fadeIn();
            attachmentElement.find("a.store_attachment").click(function(event) {
                var attachmentId = parseInt($(this).attr("attachment_id"));
                getInboxWindowController().storeAttachment(attachmentId);
                event.preventDefault();
            });
        });
    });
}

function createQuotedFromFragment(fragment) {
    var quoted = document.createElement("div");
    quoted.appendChild(fragment);
    quoted = $(quoted);
    quoted.addClass("quoted");
    quoted.find("div.toggle_quoted").remove();
    return quoted;
}

function prepareDraftQuoted(draftElement, formElement, quoted) {
    quoted.find('span.pl').click(function(event) {
        var target = $(event.target);
        splitQuoted(draftElement, formElement, quoted, target);
        event.preventDefault();
    });
}

function splitQuoted(draftElement, formElement, quoted, placeholder) {
    save_draft(draftElement, function() {
        var plid = placeholder.attr('plid');
        var body_fragment_id = quoted.parents("div.body_fragment").attr("body_fragment_id");
        var url = "/inbox/draft/" + draftElement.attr("draft_id")
            + "/body_fragment/" + body_fragment_id + "/insert_before/" + getUndoContext();
        var data = {
            placeholder_id : plid
        };
        $.post(url, data, function(result) {
            var newDraft = $(result);
            replaceDraft(draftElement, newDraft);
        });
    });
}

function focusBodyFragment(bodyFragment, setFocus) {
    $('div.body_fragment.focus').removeClass('focus'); // remove all
    bodyFragment.addClass('focus');
    if (setFocus) {
        var richtext = bodyFragment.find("section.richtext");
        richtext.focus();
    }
}

function getAbsolutePosition(element) {
    if (element.length == 0) {
        return null;
    }
    element = element.get(0); // get the real object
    if (!element) {
        return null;
    }
    var res = new Object();
    res.x = element.offsetLeft;
    res.y = element.offsetTop;

    var offsetParent = element.offsetParent;
    var parentNode = element.parentNode;

    while (offsetParent) {
        res.x += offsetParent.offsetLeft;
        res.y += offsetParent.offsetTop;

        if (offsetParent != document.body && offsetParent != document.documentElement) {
            res.x -= offsetParent.scrollLeft;
            res.y -= offsetParent.scrollTop;
        }
        parentNode = offsetParent.parentNode;
        offsetParent = offsetParent.offsetParent;
    }
    return res;
}

function getVisiblePosition(element) {
    if (!element) {
        return null;
    }
    var body = $('body');
    if (element.position().top < body.scrollTop()) {
        return null;
    }
    if (element.position().left < body.scrollLeft()) {
        return null;
    }
    var position = getAbsolutePosition(element);
    position.x -= body.scrollLeft();
    position.y -= body.scrollTop();
    return position;
}


function computeThreadPingPong() {
    var thread = $(".thread");
    if (!thread.hasClass("isdraft") && $(".messagecontainer").size() > 1) {
        thread.addClass("pingpong");
        thread.removeClass("notpingpong");
    } else {
        thread.removeClass("pingpong");
        thread.addClass("notpingpong");
    }
}

function isConversationDraft() {
    return $(".conversation.isdraft").length > 0;
}
